@@ -260,7 +260,32 @@ parse_section_urgency(const char *key, const char *value,
                       struct urgency_config *conf,
                       const char *path, unsigned lineno)
 {
-    if (strcmp(key, "background") == 0) {
+    if (strcmp(key, "layer") == 0) {
+        enum zwlr_layer_shell_v1_layer layer = 0;
+
+        if (strcasecmp(value, "background") == 0)
+            layer = ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND;
+        else if (strcasecmp(value, "top") == 0)
+            layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
+        else if (strcasecmp(value, "bottom") == 0)
+            layer = ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM;
+        else if (strcasecmp(value, "overlay") == 0)
+            layer = ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY;
+        else {
+            LOG_ERR(
+                "%s:%u: %s: invalid layer value, must be one of "
+                "\"background\", "
+                "\"bottom\", "
+                "\"top\" or "
+                "\"overlay\"",
+                path, lineno, value);
+            return false;
+        }
+
+        conf->layer = layer;
+    }
+
+    else if (strcmp(key, "background") == 0) {
         pixman_color_t bg;
         if (!str_to_color(value, &bg, path, lineno))
             return false;
@@ -495,17 +520,13 @@ parse_section_main(const char *key, const char *value, struct config *conf,
     }
 
     else if (strcmp(key, "dpi-aware") == 0) {
-        if (strcasecmp(value, "auto") == 0)
-            conf->dpi_aware = DPI_AWARE_AUTO;
-        else {
-            bool enabled;
-            if (str_to_bool(value, &enabled))
-                conf->dpi_aware = enabled ? DPI_AWARE_YES : DPI_AWARE_NO;
-            else {
-                LOG_ERR("%s:%d: %s: invalid boolean value", path, lineno, value);
-                return false;
-            }
+        bool enabled;
+        if (!str_to_bool(value, &enabled)) {
+            LOG_ERR("%s:%d: %s: invalid boolean value", path, lineno, value);
+            return false;
         }
+
+        conf->dpi_aware = enabled;
     }
 
     else if (strcmp(key, "icon-theme") == 0) {
@@ -540,14 +561,16 @@ parse_section_main(const char *key, const char *value, struct config *conf,
 
 
     else if (strcmp(key, "layer") == 0) {
+        enum zwlr_layer_shell_v1_layer layer = 0;
+
         if (strcasecmp(value, "background") == 0)
-            conf->layer = ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND;
+            layer = ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND;
         else if (strcasecmp(value, "top") == 0)
-            conf->layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
+            layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
         else if (strcasecmp(value, "bottom") == 0)
-            conf->layer = ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM;
+            layer = ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM;
         else if (strcasecmp(value, "overlay") == 0)
-            conf->layer = ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY;
+            layer = ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY;
         else {
             LOG_ERR(
                 "%s:%u: %s: invalid layer value, must be one of "
@@ -558,6 +581,9 @@ parse_section_main(const char *key, const char *value, struct config *conf,
                 path, lineno, value);
             return false;
         }
+
+        for (int i = 0; i < 3; i++)
+            conf->by_urgency[i].layer = layer;
     }
 
     else if (strcmp(key, "anchor") == 0) {
@@ -622,11 +648,49 @@ parse_section_main(const char *key, const char *value, struct config *conf,
         conf->selection_helper = strdup(value);
     }
 
+    else if (strcmp(key, "selection-helper-uses-null-separator") == 0) {
+        bool enabled;
+        if (!str_to_bool(value, &enabled)) {
+            LOG_ERR("%s:%d: %s: invalid boolean value", path, lineno, value);
+            return false;
+        }
+
+        conf->selection_helper_uses_null_separator = enabled;
+    }
+
     else if (strcmp(key, "play-sound") == 0) {
         if (!str_to_spawn_template(conf, value, &conf->play_sound, path, lineno))
             return false;
     }
 
+    else if (strcmp(key, "scaling-filter") == 0) {
+        enum scaling_filter filter;
+
+        if (strcasecmp(value, "none") == 0)
+            filter = SCALING_FILTER_NONE;
+        else if (strcasecmp(value, "nearest") == 0)
+            filter = SCALING_FILTER_NEAREST;
+        else if (strcasecmp(value, "bilinear") == 0)
+            filter = SCALING_FILTER_BILINEAR;
+        else if (strcasecmp(value, "cubic") == 0)
+            filter = SCALING_FILTER_CUBIC;
+        else if (strcasecmp(value, "lanczos3") == 0)
+            filter = SCALING_FILTER_LANCZOS3;
+        else {
+            LOG_ERR(
+                "%s:%u: %s: invalid scaling-filter value, must be one of "
+                "\"none\", "
+                "\"nearest\", "
+                "\"bilinear\", "
+                "\"cubic\" or "
+                "\"lanczos3\"",
+                path, lineno, value);
+            return false;
+        }
+
+        conf->scaling_filter = filter;
+    }
+
     else if (strcmp(key, "background") == 0) {
         pixman_color_t bg;
         if (!str_to_color(value, &bg, path, lineno))
@@ -997,12 +1061,11 @@ config_load(struct config *conf, const char *path)
         .min_width = 0,
         .max_width = 0,
         .max_height = 0,
-        .dpi_aware = DPI_AWARE_AUTO,
+        .dpi_aware = false,
         .icon_theme_name = strdup("hicolor"),
         .max_icon_size = 48,
         .stacking_order = STACK_BOTTOM_UP,
         .anchor = ANCHOR_TOP_RIGHT,
-        .layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP,
         .margins = {
             .vertical = 10,
             .horizontal = 10,
@@ -1011,6 +1074,7 @@ config_load(struct config *conf, const char *path)
         .by_urgency = {
             /* urgency == low */
             {
+                .layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP,
                 .bg = {0x2b2b, 0x2b2b, 0x2b2b, 0xffff},
                 .border = {
                     .color = {0x9090, 0x9090, 0x9090, 0xffff},
@@ -1045,6 +1109,7 @@ config_load(struct config *conf, const char *path)
                 .icon = NULL,
             },
             {
+                .layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP,
                 .bg = {0x3f3f, 0x5f5f, 0x3f3f, 0xffff},
                 .border = {
                     .color = {0x9090, 0x9090, 0x9090, 0xffff},
@@ -1079,6 +1144,7 @@ config_load(struct config *conf, const char *path)
                 .icon = NULL,
             },
             {
+                .layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP,
                 .bg = {0x6c6c, 0x3333, 0x3333, 0xffff},
                 .border = {
                     .color = {0x9090, 0x9090, 0x9090, 0xffff},
@@ -1114,6 +1180,7 @@ config_load(struct config *conf, const char *path)
             },
         },
         .selection_helper = strdup("dmenu"),
+        .scaling_filter = SCALING_FILTER_LANCZOS3,
     };
 
     for (size_t i = 0; i < sizeof(conf->by_urgency) / sizeof(conf->by_urgency[0]); i++) {
